// SNF4CGP/ScannerPool.cpp
// Copyright (C) 2009 ARM Research Labs, LLC.
// See www.armresearch.com for more information.

#include "ScannerPool.hpp"

#include "../SNFMulti/SNFMulti.hpp"
#include "../CodeDweller/threading.hpp"
#include "../CodeDweller/faults.hpp"

using namespace std;

//// Scanner ///////////////////////////////////////////////////////////////////

Scanner::Scanner(snf_RulebaseHandler& Rulebase) :                               // When constructed allocates an engine.
  Rulebase_(Rulebase),                                                          // At it's core is a Rulebase handle.
  Engine_(new snf_EngineHandler()) {                                            // It contains it's own scan engine.
    Engine_->open(&Rulebase);                                                   // The scan engine is initialized.
}

Scanner::~Scanner() {                                                           // When destructed destroys the engine.
    try {                                                                       // Destructors don't throw.
        if(Engine_) {                                                           // Safely destroy the engine.
            Engine_->close();
            delete Engine_;
            Engine_=0;
        }
    }
    catch(...) {}
}

snf_RulebaseHandler& Scanner::Rulebase() {                                      // Provide access to the Rulebase
    return Rulebase_;
}

RuntimeCheck CheckScannerEngineIsValid("Scanner::Engine() Check(0 != Engine_)");

snf_EngineHandler& Scanner::Engine() {                                          // Provide access to the Engine.
    CheckScannerEngineIsValid(0 != Engine_);
    return (*Engine_);
}

//// ScannerPool ///////////////////////////////////////////////////////////////

ScannerPool::ScannerPool() :                                                    // Constructed simply.
  Rulebase_(new snf_RulebaseHandler()),
  ScannerConfiguration_(new ConfigurationManager(Rulebase())),
  ScannerCount(0),
  Started(false) {
}

ScannerPool::~ScannerPool() {                                                   // Cleanup on the way out.
    try {
        stop();
        delete Rulebase_;
        Rulebase_ = 0;
    }
    catch(...) {}
}

RuntimeCheck CheckRulebaseExists("ScannerPool::Rulebase() Check(0 != Rulebase_)");

snf_RulebaseHandler& ScannerPool::Rulebase() {                                  // Safe access to the rulebase.
    CheckRulebaseExists(0 != Rulebase_);
    return (*Rulebase_);
}

void ScannerPool::init(const string VersionInfo, const string Configuration) {  // Get ready to provide scanners.
    Rulebase().PlatformVersion(VersionInfo);
    Rulebase().open(Configuration.c_str(), "", "");                             // Light up the rulebase manager.
    ScopeMutex Busy(AllocationControl);
    Scanner* FirstScanner = makeScanner();                                      // Create our first scanner.
    PooledScanners.give(FirstScanner);                                          // Drop it into the pool.
    Started = true;
}

void ScannerPool::killScannerFromPool() {                                       // Destroy and count a pooled scanner.
    Scanner* ScannerToKill = 0;
    ScannerToKill = PooledScanners.take();
    delete ScannerToKill;
    --ScannerCount;
}

void ScannerPool::stop() {                                                      // Shutdown and clean up.
    ScopeMutex Busy(AllocationControl);
    while(0 < ScannerCount) killScannerFromPool();
    Rulebase().close();
    Started = false;
}

Scanner* ScannerPool::makeScanner() {                                           // Create and count a scanner.
    Scanner* NewScanner = new Scanner(Rulebase());
    ++ScannerCount;
    return NewScanner;
}

LogicCheck CheckGrabScannersOnlyWhenStarted("ScannerPool::grab() Check(Started)");
RuntimeCheck CheckForValidGrabbedScanner("ScannerPool::grab() Check(0 != GrabbedScanner)");

Scanner& ScannerPool::grab() {                                                  // Provides a Scanner from the pool.
    ScopeMutex Busy(AllocationControl);
    CheckGrabScannersOnlyWhenStarted(Started);
    Scanner* GrabbedScanner = 0;
    if(0 < PooledScanners.size()) GrabbedScanner = PooledScanners.take();       // Prefer scanners from the pool.
    else GrabbedScanner = makeScanner();                                        // Make new ones if needed.
    CheckForValidGrabbedScanner(0 != GrabbedScanner);
    return (*GrabbedScanner);
}

void ScannerPool::drop(Scanner& S) {                                            // Returns a Scanner to the pool.
    PooledScanners.give(&S);
}

RuntimeCheck CheckForValidConfigurationManager("ScannerPool::ScannerConfiguration() Check(0 != ScannerConfiguration_)");

ConfigurationManager& ScannerPool::ScannerConfiguration() {                     // Safe coniguration access & update logic.
    CheckForValidConfigurationManager(0 != ScannerConfiguration_);
    ScopeMutex CheckingConfiguration(ConfigurationMutex);
    ConfigurationManager& TheConfiguration = (*ScannerConfiguration_);
    if(TheConfiguration.isOutOfDate()) TheConfiguration.update();
    return TheConfiguration;
}

ResultConfiguration ScannerPool::ConfigurationForResultCode(int Code) {         // Multiplexed access to configuration.
    return ScannerConfiguration().ConfigurationForResultCode(Code);
}

//// ScopeScanner //////////////////////////////////////////////////////////////

ScopeScanner::ScopeScanner(ScannerPool& P) :                                    // Constructed with the pool.
  ScannerPool_(P),
  Scanner_(P.grab()) {                                                          // Grabs a Scanner on the way up.
}

ScopeScanner::~ScopeScanner() {                                                 // When destroyed drops the scanner.
    try { ScannerPool_.drop(Scanner_); }
    catch(...) {}
}

snf_RulebaseHandler& ScopeScanner::Rulebase() {                                 // Provides access to the Rulebase.
    return Scanner_.Rulebase();
}

snf_EngineHandler& ScopeScanner::Engine() {                                     // Provides access to a scan engine.
    return Scanner_.Engine();
}
